@@ -9,7 +9,8 @@ import click
 import tzlocal
 import logging
 
-from .config_utils import pyca_cryptography_present, ConfigurationError
+from .config_utils import ConfigurationError
+from .crypto_utils import pyca_cryptography_present
 from .registry import CertomancerConfig, CertLabel
 from .services import CertomancerServiceError
 from .version import __version__
@@ -48,12 +49,14 @@ def exception_manager():
         raise click.ClickException(msg)
 
 
-def _lazy_cfg(config, key_root, cfg_root, no_external_config):
+def _lazy_cfg(config, key_root, cfg_root, no_external_config,
+              service_url_prefix):
     config = config or DEFAULT_CONFIG_FILE
     try:
         cfg = CertomancerConfig.from_file(
             config, key_search_dir=key_root, config_search_dir=cfg_root,
-            allow_external_config=not no_external_config
+            allow_external_config=not no_external_config,
+            external_url_prefix=service_url_prefix
         )
     except IOError as e:
         raise click.ClickException(
@@ -80,13 +83,18 @@ def _lazy_cfg(config, key_root, cfg_root, no_external_config):
               required=False, type=click.Path(readable=True, file_okay=False))
 @click.option('--no-external-config', help='disable external config loading',
               required=False, type=bool, is_flag=True)
+@click.option('--service-url-prefix',
+              help='override configured URL prefix for service URLs',
+              required=False, type=str)
 @click.pass_context
 @exception_manager()
-def cli(ctx, config, key_root, extra_config_root, no_external_config):
+def cli(ctx, config, key_root, extra_config_root, no_external_config,
+        service_url_prefix):
     _log_config()
     ctx.ensure_object(dict)
     ctx.obj['config'] = _lazy_cfg(
-        config, key_root, extra_config_root, no_external_config
+        config, key_root, extra_config_root, no_external_config,
+        service_url_prefix
     )
 
 
@@ -233,14 +241,21 @@ def necronomicon(ctx, architecture, crl_repo, output, no_pem, at_time,
               required=False, type=bool, is_flag=True)
 @click.option('--no-time-override', help='disable time override functionality',
               required=False, type=bool, is_flag=True)
+@click.option('--wsgi-prefix', required=False, type=str,
+              help=(
+                      'WSGI prefix under which to mount the application '
+                      '(does not affect generated output)'
+              ))
 @click.pass_context
 @exception_manager()
-def animate(ctx, port, no_web_ui, no_time_override):
+def animate(ctx, port, no_web_ui, no_time_override, wsgi_prefix):
     try:
         from .integrations.animator import Animator, AnimatorArchStore
+        from werkzeug.middleware.dispatcher import DispatcherMiddleware
+        from werkzeug.exceptions import NotFound
     except ImportError as e:
         raise click.ClickException(
-            "'animate' requires additional dependencies."
+            "'animate' requires additional dependencies. "
             "Re-run setup with the [web-api] extension set, or install "
             "Werkzeug (and Jinja2 for the web UI) manually."
         ) from e
@@ -250,4 +265,9 @@ def animate(ctx, port, no_web_ui, no_time_override):
         AnimatorArchStore(cfg.pki_archs), with_web_ui=not no_web_ui,
         allow_time_override=not no_time_override
     )
+    if wsgi_prefix:
+        # Serve the Animator under the indicated prefix, wrapped using
+        # dispatcher middleware (functionally equivalent to SCRIPT_NAME, but
+        # more convenient to run from the CLI in this way)
+        app = DispatcherMiddleware(NotFound(), {wsgi_prefix: app})
     run_simple('127.0.0.1', port, app)
