@@ -1,4 +1,5 @@
 import abc
+import copy
 import hashlib
 import importlib
 import logging
@@ -16,13 +17,17 @@ from asn1crypto.core import ObjectIdentifier
 from dateutil.parser import parse as parse_dt
 from asn1crypto import x509, core, pem, ocsp, crl
 from dateutil.tz import tzlocal
-from oscrypto import keys as oskeys, asymmetric
 from asn1crypto.keys import PrivateKeyInfo, PublicKeyInfo
 
 from .config_utils import (
     ConfigurationError, check_config_keys, LabelString,
     ConfigurableMixin, parse_duration, key_dashes_to_underscores, get_and_apply,
-    pyca_cryptography_present, SearchDir, plugin_instantiate_util
+    SearchDir, plugin_instantiate_util
+)
+from .crypto_utils import (
+    pyca_cryptography_present,
+    load_public_key, load_private_key,
+    load_cert_from_pemder
 )
 from .services import CertomancerServiceError, generic_sign, CRLBuilder, \
     choose_signed_digest, SimpleOCSPResponder, TimeStamper, \
@@ -131,13 +136,10 @@ class KeyFromFile:
             with open(self.path, 'rb') as keyf:
                 key_bytes = keyf.read()
             if self.public_only:
-                public = oskeys.parse_public(key_bytes)
+                public = load_public_key(key_bytes)
                 private = None
             else:
-                private = oskeys.parse_private(
-                    key_bytes, password=self.password
-                )
-                public = asymmetric.load_private_key(private).public_key.asn1
+                private, public = load_private_key(key_bytes, self.password)
             self._key = AsymKey(public=public, private=private)
 
     @property
@@ -160,7 +162,9 @@ class KeySet:
     """A labelled collection of keys."""
 
     def __init__(self, config, search_dir: SearchDir, lazy_load_keys=False):
-        check_config_keys('KeySet', ('path-prefix', 'keys'), config)
+        check_config_keys(
+            'KeySet', ('path-prefix', 'keys', 'default-password'), config
+        )
         try:
             keys = config['keys']
         except KeyError as e:
@@ -170,17 +174,22 @@ class KeySet:
         path_prefix = config.get('path-prefix', '')
         path_prefix = search_dir.search_subdir(path_prefix)
 
+        default_password = config.get('default-password', None)
+
         # apply path prefix to key configs
-        def _prepend(key_conf):
+        def _proc(key_conf):
+            key_conf = copy.deepcopy(key_conf)
             try:
                 key_conf['path'] = path_prefix.resolve(key_conf['path'])
             except KeyError:
                 pass
+            if default_password is not None:
+                key_conf.setdefault('password', default_password)
             return key_conf
 
         self._dict = {
             KeyLabel(k): KeyFromFile.from_config(
-                KeyLabel(k), _prepend(v), lazy=lazy_load_keys
+                KeyLabel(k), _proc(v), lazy=lazy_load_keys
             )
             for k, v in keys.items()
         }
@@ -212,10 +221,36 @@ class KeySets:
     """A labelled collection of key sets."""
 
     def __init__(self, config, search_dir, lazy_load_keys=False):
-        self._dict = {
-            k: KeySet(v, lazy_load_keys=lazy_load_keys, search_dir=search_dir)
-            for k, v in config.items()
-        }
+        results = {}
+        configs_seen = {}
+        for k, cfg in config.items():
+            cfg = copy.deepcopy(cfg)
+            if 'template' in cfg:
+                template_keyset = cfg.pop('template')
+                try:
+                    template_cfg = copy.deepcopy(configs_seen[template_keyset])
+                except KeyError as e:
+                    raise ConfigurationError(
+                        f"Key set definition with label '{k}' "
+                        f"refers to '{template_keyset}' as a template, but "
+                        f"'{template_keyset}' hasn't been declared yet."
+                    ) from e
+                # merge 'keys' entries
+                template_keys = template_cfg['keys']
+                try:
+                    extra_keys = cfg.pop('keys')
+                    template_keys.update(extra_keys)
+                except KeyError:
+                    pass
+                # clobber all other entries
+                template_cfg.update(cfg)
+                # replace cfg with the updated one
+                cfg = template_cfg
+            configs_seen[k] = cfg
+            results[k] = KeySet(
+                cfg, lazy_load_keys=lazy_load_keys, search_dir=search_dir
+            )
+        self._dict = results
 
     def __getitem__(self, name) -> KeySet:
         try:
@@ -302,6 +337,17 @@ class Validity(ConfigurableMixin):
                 "Illegal date-time string in validity specification"
             ) from e
 
+    @property
+    def asn1(self) -> x509.Validity:
+        return x509.Validity({
+            'not_before': x509.Time(
+                {'utc_time' if self.valid_from.year < 2050 else 'general_time': self.valid_from}
+            ),
+            'not_after': x509.Time(
+                {'utc_time' if self.valid_to.year < 2050 else 'general_time': self.valid_to}
+            ),
+        })
+
 
 class ExtensionPlugin(abc.ABC):
     """
@@ -418,7 +464,14 @@ class ExtensionPluginRegistry:
             extn_id = proc.extension_type(extn_id)
         else:
             extn_id = None
-        return proc.provision(extn_id, arch, spec.params)
+        provisioned_value = proc.provision(extn_id, arch, spec.params)
+        if isinstance(provisioned_value, core.Asn1Value) and \
+                not isinstance(provisioned_value, core.ParsableOctetString):
+            # this allows plugins to keep working with extensions for which
+            # we don't have an OID
+            provisioned_value = \
+                core.ParsableOctetString(provisioned_value.dump())
+        return provisioned_value
 
 
 DEFAULT_EXT_PLUGIN_REGISTRY = extension_plugin_registry \
@@ -486,7 +539,9 @@ class ExtensionSpec(ConfigurableMixin):
         })
 
 
-EXCLUDED_FROM_TEMPLATE = frozenset({'subject', 'subject_key', 'serial'})
+EXCLUDED_FROM_TEMPLATE = frozenset(
+    {'subject', 'subject_key', 'serial', 'certificate_file'}
+)
 EXTNS_EXCLUDED_FROM_TEMPLATE = frozenset({'subject_alt_name'})
 
 
@@ -619,6 +674,27 @@ class CertificateSpec(ConfigurableMixin):
         specified in the configuration.
     """
 
+    certificate_file: Optional[str] = None
+    """
+    Path to a file with a pre-generated copy of the certificate in question,
+    either in DER or in PEM format.
+    
+    When the certificate determined by this certificate spec is requested,
+    the certificate in the file will be returned.
+    
+    .. warning::
+        Certomancer will not attempt to process any information from the
+        certificate file, beyond parsing it into an X.509 certificate structure.
+        Internally, the certificate spec's entries are used instead.
+        It is the config writer's responsibility to make sure that both match
+        up.
+
+    .. note::
+        This option is unavailable when external configuration is disabled.
+        Moreover, it is excluded from templates derived from this certificate
+        spec.
+    """
+
     @property
     def self_issued(self) -> bool:
         """
@@ -698,15 +774,16 @@ class PKIArchitecture:
     modelled by Certomancer.
     """
 
-    CONFIG_KEYS = (
-        'keyset', 'entities', 'certs', 'services', 'entity-defaults'
-    )
+    # These config keys will be merged when an architecture is templated
+    MULTIVAL_CONFIG_KEYS = ('entities', 'certs', 'services', 'entity-defaults')
+    CONFIG_KEYS = ('keyset', *MULTIVAL_CONFIG_KEYS)
 
     @classmethod
     def build_architecture(cls, arch_label: ArchLabel, cfg: dict,
                            key_sets: KeySets, external_url_prefix,
                            extension_plugins: ExtensionPluginRegistry = None,
                            service_plugins: 'ServicePluginRegistry' = None,
+                           config_search_dir: Optional[SearchDir] = None,
                            cert_cache=None) -> 'PKIArchitecture':
         check_config_keys(arch_label, PKIArchitecture.CONFIG_KEYS, cfg)
         key_set_label = cfg.get('keyset', arch_label)
@@ -740,6 +817,7 @@ class PKIArchitecture:
             cert_spec_config=cert_specs, service_config=services,
             external_url_prefix=external_url_prefix,
             extension_plugins=extension_plugins,
+            config_search_dir=config_search_dir,
             service_plugins=service_plugins, cert_cache=cert_cache
         )
 
@@ -748,6 +826,7 @@ class PKIArchitecture:
                             config_search_dir: Optional[SearchDir],
                             extension_plugins: ExtensionPluginRegistry = None,
                             service_plugins: 'ServicePluginRegistry' = None):
+        arch_specs = {}
         for arch_label, cfg in cfgs.items():
             arch_label = ArchLabel(arch_label)
             # external config
@@ -760,10 +839,52 @@ class PKIArchitecture:
                 cfg_path = config_search_dir.resolve(path=cfg)
                 with open(cfg_path, 'r') as external_conf:
                     cfg = yaml.safe_load(external_conf)
+            elif isinstance(cfg, dict):
+                # make sure we don't mess with input data passed in
+                # by the caller
+                cfg = copy.deepcopy(cfg)
+            else:
+                raise ConfigurationError(
+                    f"Architecture definition must be either a string or a "
+                    f"dictionary; config with label {arch_label} has type "
+                    f"{type(cfg)}."
+                )
+            if 'template' in cfg:
+                template_arch = cfg.pop('template')
+                # retrieve template config
+                try:
+                    template_cfg = copy.deepcopy(arch_specs[template_arch])
+                except KeyError as e:
+                    raise ConfigurationError(
+                        f"Architecture definition with label '{arch_label}' "
+                        f"refers to '{template_arch}' as a template, but "
+                        f"'{template_arch}' hasn't been declared yet."
+                    ) from e
+
+                # first, merge multivalue config keys
+                for key in PKIArchitecture.MULTIVAL_CONFIG_KEYS:
+                    try:
+                        extra_values = cfg.pop(key)
+                    except KeyError:
+                        continue
+                    try:
+                        orig_values = template_cfg[key]
+                    except KeyError:
+                        template_cfg[key] = orig_values = {}
+                    # update effective config with values to merge
+                    orig_values.update(extra_values)
+
+                # then clobber the rest (all the multivalued keys should have
+                # been deleted by now)
+                template_cfg.update(cfg)
+                cfg = template_cfg
+            # store the config for potential later template use
+            arch_specs[arch_label] = cfg
             yield cls.build_architecture(
                 arch_label=arch_label, cfg=cfg, key_sets=key_sets,
                 external_url_prefix=external_url_prefix,
                 extension_plugins=extension_plugins,
+                config_search_dir=config_search_dir,
                 service_plugins=service_plugins
             )
 
@@ -772,6 +893,7 @@ class PKIArchitecture:
                  cert_spec_config, service_config, external_url_prefix,
                  extension_plugins: ExtensionPluginRegistry = None,
                  service_plugins: 'ServicePluginRegistry' = None,
+                 config_search_dir: Optional[SearchDir] = None,
                  cert_cache=None):
 
         self.arch_label = arch_label
@@ -796,6 +918,8 @@ class PKIArchitecture:
         self._cert_labels_by_subject: Dict[EntityLabel, List[CertLabel]] \
             = defaultdict(list)
 
+        self._cert_cache = cert_cache if cert_cache is not None else {}
+
         serial_by_issuer = defaultdict(lambda: DEFAULT_FIRST_SERIAL)
         for name, cert_config in cert_spec_config.items():
             name = CertLabel(name)
@@ -841,11 +965,24 @@ class PKIArchitecture:
             cert_specs[name] = spec = CertificateSpec.from_config(
                 effective_cert_config
             )
+            if spec.certificate_file is not None:
+                if config_search_dir is None:
+                    raise ConfigurationError(
+                        f"Failed to load pregenerated cert with name {name}"
+                        f"from file; external configuration is disabled."
+                    )
+                full_path = config_search_dir.resolve(spec.certificate_file)
+                try:
+                    pregenerated = load_cert_from_pemder(full_path)
+                    self._cert_cache[name] = pregenerated
+                except (IOError, ValueError) as e:
+                    raise ConfigurationError(
+                        f"Failed to load pregenerated cert with name {name}"
+                        f"from file {full_path}."
+                    ) from e
             self._cert_labels_by_issuer[spec.issuer].append(name)
             self._cert_labels_by_subject[spec.subject].append(name)
 
-        self._cert_cache = cert_cache if cert_cache is not None else {}
-
     def get_cert_spec(self, label: CertLabel) -> CertificateSpec:
         try:
             return self._cert_specs[label]
@@ -1025,7 +1162,7 @@ class PKIArchitecture:
         signature_algo = spec.signature_algo
         digest_algo = spec.digest_algo
         signature_algo_obj = choose_signed_digest(
-            digest_algo, authority_key.public_key_info.algorithm,
+            digest_algo, authority_key.public_key_info,
             signature_algo
         )
 
@@ -1068,14 +1205,7 @@ class PKIArchitecture:
             'serial_number': spec.serial,
             'signature': signature_algo_obj,
             'issuer': issuer_name,
-            'validity': x509.Validity({
-                'not_before': x509.Time(
-                    {'general_time': spec.validity.valid_from}
-                ),
-                'not_after': x509.Time(
-                    {'general_time': spec.validity.valid_to}
-                ),
-            }),
+            'validity': spec.validity.asn1,
             'subject': subject_name,
             'subject_public_key_info': subject_key.public_key_info,
             'extensions': extensions
@@ -1159,13 +1289,24 @@ class ServiceInfo(ConfigurableMixin):
 
         return f"{self.base_url}/{self.label}"
 
+    @property
+    def full_relative_url(self):
+        """
+        Full URL where the service's main endpoint can be found,
+        relative to :attr:`external_url_prefix`.
+
+        This is the URL used when listing service links in the web UI.
+        """
+        return f"{self.arch_label}{self.internal_url}"
+
     @property
     def url(self) -> str:
         """
         Full URL where the service's main endpoint can be found.
+
+        This is the value that is embedded into certificates.
         """
-        return \
-            f"{self.external_url_prefix}/{self.arch_label}{self.internal_url}"
+        return f"{self.external_url_prefix}/{self.full_relative_url}"
 
 
 @dataclass(frozen=True)
@@ -1367,14 +1508,14 @@ class CRLRepoServiceInfo(ServiceInfo):
 
         _parse_extension_settings(config_dict, 'crl_extensions')
 
-    @property
-    def latest_url(self):
-        return f"{self.internal_url}/latest.crl"
-
     @property
     def latest_external_url(self):
         return f"{self.url}/latest.crl"
 
+    @property
+    def latest_full_relative_url(self):
+        return f"{self.full_relative_url}/latest.crl"
+
     def archive_url(self, for_crl_number):
         return f"{self.internal_url}/archive-{for_crl_number}.crl"
 
@@ -1413,9 +1554,14 @@ class CertRepoServiceInfo(ServiceInfo):
         return f"{self.url}/{fname}"
 
     @property
-    def issuer_cert_internal_url(self):
+    def issuer_cert_external_url(self):
         fname = CertRepoServiceInfo.issuer_cert_file_name(use_pem=True)
-        return f"{self.internal_url}/{fname}"
+        return f"{self.url}/{fname}"
+
+    @property
+    def issuer_cert_full_relative_url(self):
+        fname = CertRepoServiceInfo.issuer_cert_file_name(use_pem=True)
+        return f"{self.full_relative_url}/{fname}"
 
     def issued_cert_url(self, label: CertLabel, use_pem=True):
         path = self.issued_cert_url_path(label=label, use_pem=use_pem)
@@ -1721,11 +1867,12 @@ class ServiceRegistry:
             ext.to_asn1(self.pki_arch, ocsp.ResponseDataExtension)
             for ext in info.ocsp_extensions
         ]
+        responder_cert = self.pki_arch.get_cert(info.responder_cert)
         return SimpleOCSPResponder(
-            responder_cert=self.pki_arch.get_cert(info.responder_cert),
+            responder_cert=responder_cert,
             responder_key=responder_key,
             signature_algo=choose_signed_digest(
-                info.digest_algo, responder_key.algorithm,
+                info.digest_algo, responder_cert.public_key,
                 signature_algo=info.signature_algo
             ),
             at_time=at_time,
@@ -1776,12 +1923,13 @@ class ServiceRegistry:
         # TODO allow policy parameter to be customised
         info = self.get_tsa_info(label)
         tsa_key = self.pki_arch.key_set.get_private_key(info.signing_key)
+        tsa_cert = self.pki_arch.get_cert(info.signing_cert)
         return TimeStamper(
-            tsa_cert=self.pki_arch.get_cert(info.signing_cert),
+            tsa_cert=tsa_cert,
             tsa_key=tsa_key,
             fixed_dt=at_time,
             signature_algo=choose_signed_digest(
-                info.digest_algo, key_algo=tsa_key.algorithm,
+                info.digest_algo, pub_key=tsa_cert.public_key,
                 signature_algo=info.signature_algo
             ),
             certs_to_embed=[
@@ -1797,8 +1945,9 @@ class ServiceRegistry:
 
         crl_info = self.get_crl_repo_info(repo_label)
         issuer_cert_label = crl_info.issuer_cert
-        signing_key = self.pki_arch.key_set.\
-            get_private_key(crl_info.signing_key)
+        signing_key_pair = \
+            self.pki_arch.key_set.get_asym_key(crl_info.signing_key)
+        signing_key = signing_key_pair.private
 
         # we need a cert to compute the right authority key identifier,
         # time origin etc.
@@ -1833,7 +1982,7 @@ class ServiceRegistry:
             issuer_name=self.pki_arch.entities[crl_info.for_issuer],
             issuer_key=signing_key,
             signature_algo=choose_signed_digest(
-                crl_info.digest_algo, signing_key.algorithm,
+                crl_info.digest_algo, signing_key_pair.public,
                 signature_algo=crl_info.signature_algo
             ),
             authority_key_identifier=iss_cert.key_identifier_value,
@@ -1936,19 +2085,23 @@ class CertomancerConfig:
     """
 
     DEFAULT_EXTERNAL_URL_PREFIX = 'http://ca.example.com'
+    EXTERNAL_URL_PREFIX_VARIABLE = 'external-url-prefix'
 
     @classmethod
     def from_yaml(cls, yaml_str, key_search_dir,
-                  config_search_dir=None) -> 'CertomancerConfig':
+                  config_search_dir=None,
+                  external_url_prefix=None) -> 'CertomancerConfig':
         config_dict = yaml.safe_load(yaml_str)
         return CertomancerConfig(
             config_dict, key_search_dir=key_search_dir,
-            config_search_dir=config_search_dir
+            config_search_dir=config_search_dir,
+            external_url_prefix=external_url_prefix
         )
 
     @classmethod
     def from_file(cls, cfg_path, key_search_dir=None, config_search_dir=None,
-                  allow_external_config=True) -> 'CertomancerConfig':
+                  allow_external_config=True,
+                  external_url_prefix=None) -> 'CertomancerConfig':
         main_config_dir = os.path.dirname(cfg_path)
         if not allow_external_config:
             config_search_dir = None
@@ -1959,14 +2112,17 @@ class CertomancerConfig:
             config_dict = yaml.safe_load(inf)
         return CertomancerConfig(
             config_dict, key_search_dir=key_search_dir,
-            config_search_dir=config_search_dir
+            config_search_dir=config_search_dir,
+            external_url_prefix=external_url_prefix
         )
 
     def __init__(self, config, key_search_dir: str,
-                 lazy_load_keys=False, config_search_dir: Optional[str] = None):
-        self.external_url_prefix = external_url_prefix = config.get(
-            'external-url-prefix', self.DEFAULT_EXTERNAL_URL_PREFIX
-        )
+                 lazy_load_keys=False, config_search_dir: Optional[str] = None,
+                 external_url_prefix=None):
+        if external_url_prefix is None:
+            self.external_url_prefix = external_url_prefix = config.get(
+                'external-url-prefix', self.DEFAULT_EXTERNAL_URL_PREFIX
+            )
 
         extn_plugin_list = config.get('plugin-modules', ())
         _import_plugin_modules(extn_plugin_list)
