@@ -7,10 +7,11 @@ from asn1crypto.crl import TBSCertListExtension
 from datetime import datetime, timedelta
 from typing import Optional, List, Tuple
 
-from oscrypto import asymmetric
 from asn1crypto import keys, x509, tsp, algos, cms, core, crl, ocsp
 import tzlocal
 
+from certomancer.crypto_utils import generic_sign, optimal_pss_params
+
 
 class CertomancerServiceError(Exception):
     pass
@@ -59,7 +60,7 @@ class TimeStamper:
         self.policy = policy
         self.md_algorithm = md_algorithm
         self.signature_algo = signature_algo or choose_signed_digest(
-            md_algorithm, tsa_key.algorithm
+            md_algorithm, tsa_cert.public_key
         )
 
     def request_tsa_response(self, req: tsp.TimeStampReq) -> tsp.TimeStampResp:
@@ -242,62 +243,40 @@ def url_distribution_point(url, extra_urls=()):
     }
 
 
-def choose_signed_digest(digest_algo: str, key_algo: str,
+def choose_signed_digest(digest_algo: str, pub_key: keys.PublicKeyInfo,
                          signature_algo: Optional[str] = None):
+    key_algo = pub_key.algorithm
     if signature_algo is None:
+        # special OID for keys that should only be used with PSS
+        if key_algo == 'rsassa_pss':
+            signature_algo = 'rsassa_pss'
         if key_algo == 'rsa':
             signature_algo = digest_algo + '_rsa'
         elif key_algo == 'dsa':
             signature_algo = digest_algo + '_dsa'
         elif key_algo == 'ec':
             signature_algo = digest_algo + '_ecdsa'
+        elif key_algo == 'ed25519':
+            signature_algo = 'ed25519'
+        elif key_algo == 'ed448':
+            signature_algo = 'ed448'
 
     signature_algo_obj = algos.SignedDigestAlgorithm(
         {'algorithm': signature_algo}
     )
     if signature_algo == 'rsassa_pss':
-        # replicate default oscrypto PSS settings
-        salt_len = len(getattr(hashlib, digest_algo)().digest())
-        signature_algo_obj['parameters'] = algos.RSASSAPSSParams({
-            'hash_algorithm': algos.DigestAlgorithm({
-                'algorithm': digest_algo
-            }),
-            'mask_gen_algorithm': algos.MaskGenAlgorithm({
-                'algorithm': 'mgf1',
-                'parameters': algos.DigestAlgorithm(
-                    {'algorithm': digest_algo}
-                )
-            }),
-            'salt_length': salt_len
-        })
+        parameters = None
+        if pub_key.algorithm == 'rsassa_pss':
+            key_params = pub_key['algorithm']['parameters']
+            if key_params.native is not None:
+                parameters = key_params
+        if parameters is None:
+            parameters = optimal_pss_params(pub_key, digest_algo)
+        signature_algo_obj['parameters'] = parameters
 
     return signature_algo_obj
 
 
-def generic_sign(private_key: keys.PrivateKeyInfo, tbs_bytes: bytes,
-                 signature_algo: algos.SignedDigestAlgorithm) -> bytes:
-
-    pk_algo = private_key.algorithm
-    if pk_algo == 'rsa':
-        if signature_algo.signature_algo == 'rsassa_pss':
-            sign_fun = asymmetric.rsa_pss_sign
-        else:
-            sign_fun = asymmetric.rsa_pkcs1v15_sign
-    elif pk_algo == 'ec':
-        sign_fun = asymmetric.ecdsa_sign
-    elif pk_algo == 'dsa':
-        sign_fun = asymmetric.dsa_sign
-    else:
-        raise NotImplementedError(
-            f"The cryptosystem '{pk_algo}' is not supported."
-        )
-
-    return sign_fun(
-        asymmetric.load_private_key(private_key),
-        tbs_bytes, signature_algo.hash_algo
-    )
-
-
 def issuer_match(cid: ocsp.CertId, candidate: x509.Certificate):
     # FIXME this doesn't really scale
     hash_algo = cid['hash_algorithm']['algorithm'].native
